@@ -32,6 +32,7 @@ from amgi_types import MessageReceiveEvent
 from amgi_types import MessageScope
 from amgi_types import MessageSendEvent
 from amgi_types import Scope
+from asyncfast.bindings import Binding
 from pydantic import BaseModel
 from pydantic import create_model
 from pydantic import TypeAdapter
@@ -43,7 +44,6 @@ from pydantic_core import CoreSchema
 from typing_extensions import get_args
 from typing_extensions import get_origin
 
-
 DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])
 
 
@@ -57,6 +57,7 @@ class Message(Mapping[str, Any]):
     __headers__: ClassVar[dict[str, TypeAdapter[Any]]]
     __parameters__: ClassVar[dict[str, TypeAdapter[Any]]]
     __payload__: ClassVar[Optional[tuple[str, TypeAdapter[Any]]]]
+    __bindings__: ClassVar[dict[str, TypeAdapter[Any]]]
 
     def __init_subclass__(cls, address: Optional[str] = None, **kwargs: Any) -> None:
         cls.__address__ = address
@@ -74,6 +75,12 @@ class Message(Mapping[str, Any]):
             if isinstance(get_args(annotated)[1], Parameter)
         }
 
+        bindings = {
+            name: TypeAdapter(annotated)
+            for name, annotated in annotations
+            if isinstance(get_args(annotated)[1], Binding)
+        }
+
         payloads = [
             (name, TypeAdapter(annotated))
             for name, annotated in annotations
@@ -87,21 +94,30 @@ class Message(Mapping[str, Any]):
         cls.__headers__ = headers
         cls.__parameters__ = parameters
         cls.__payload__ = payload
+        cls.__bindings__ = bindings
 
     def __getitem__(self, key: str, /) -> Any:
         if key == "address":
             return self._get_address()
         elif key == "headers":
             return self._get_headers()
-        elif key == "payload":
+        elif key == "payload" and self.__payload__:
             return self._get_payload()
+        elif key == "bindings" and self.__bindings__:
+            return self._get_bindings()
         raise KeyError(key)
 
     def __len__(self) -> int:
-        return 3
+        payload = 1 if self.__payload__ else 0
+        bindings = 1 if self.__bindings__ else 0
+        return 2 + payload + bindings
 
     def __iter__(self) -> Iterator[str]:
-        return iter(("address", "headers", "payload"))
+        yield from ("address", "headers")
+        if self.__payload__:
+            yield "payload"
+        if self.__bindings__:
+            yield "bindings"
 
     def _get_address(self) -> Optional[str]:
         if self.__address__ is None:
@@ -132,6 +148,17 @@ class Message(Mapping[str, Any]):
         name, type_adapter = self.__payload__
         return type_adapter.dump_json(getattr(self, name))
 
+    def _get_bindings(self) -> dict[str, dict[str, Any]]:
+        bindings: dict[str, dict[str, Any]] = {}
+        for name, type_adapter in self.__bindings__.items():
+            binding_type = get_args(type_adapter._type)[1]
+            assert isinstance(binding_type, Binding)
+
+            bindings.setdefault(binding_type.__protocol__, {})[
+                binding_type.__field_name__
+            ] = self._get_value(name, type_adapter)
+        return bindings
+
 
 def _generate_message_annotations(
     address: Optional[str],
@@ -213,6 +240,12 @@ class AsyncFast:
             if isinstance(get_args(annotated)[1], Payload)
         ]
 
+        bindings = {
+            name: TypeAdapter(annotated)
+            for name, annotated in annotations
+            if isinstance(get_args(annotated)[1], Binding)
+        }
+
         assert len(payloads) <= 1, "Channel must have no more than 1 payload"
 
         payload = payloads[0] if len(payloads) == 1 else None
@@ -220,7 +253,14 @@ class AsyncFast:
         address_pattern = _address_pattern(address)
 
         channel = Channel(
-            address, address_pattern, function, headers, parameters, payload, messages
+            address,
+            address_pattern,
+            function,
+            headers,
+            parameters,
+            payload,
+            messages,
+            bindings,
         )
 
         self._channels.append(channel)
@@ -277,6 +317,11 @@ class AsyncFast:
         self,
     ) -> Generator[tuple[int, JsonSchemaMode, CoreSchema], None, None]:
         for channel in self._channels:
+            for type_adapter in channel._bindings.values():
+                yield hash(
+                    type_adapter._type
+                ), "serialization", type_adapter.core_schema
+
             headers_model = channel.headers_model
             if headers_model:
                 yield hash(headers_model), "serialization", TypeAdapter(
@@ -297,6 +342,11 @@ class AsyncFast:
                         type_adapter._type
                     ), "serialization", type_adapter.core_schema
 
+                for type_adapter in message.__bindings__.values():
+                    yield hash(
+                        type_adapter._type
+                    ), "serialization", type_adapter.core_schema
+
 
 def _generate_annotations(
     address: str,
@@ -393,6 +443,7 @@ class Channel:
         parameters: Mapping[str, TypeAdapter[Any]],
         payload: Optional[tuple[str, TypeAdapter[Any]]],
         messages: Sequence[type[Message]],
+        bindings: Mapping[str, TypeAdapter[Any]],
     ) -> None:
         self._address = address
         self._address_pattern = address_pattern
@@ -401,6 +452,7 @@ class Channel:
         self._parameters = parameters
         self._payload = payload
         self._messages = messages
+        self._bindings = bindings
 
     @property
     def address(self) -> str:
@@ -518,6 +570,18 @@ class Channel:
             for name, type_adapter in self._parameters.items():
                 yield name, type_adapter.validate_python(parameters[name])
 
+        if self._bindings:
+            bindings = message_receive_event.get("bindings", {})
+            for name, type_adapter in self._bindings.items():
+                binding_type = get_args(type_adapter._type)[1]
+                assert isinstance(binding_type, Binding)
+
+                yield name, type_adapter.validate_python(
+                    bindings.get(binding_type.__protocol__, {}).get(
+                        binding_type.__field_name__
+                    )
+                )
+
 
 def _generate_messages(
     channels: Iterable[Channel],
@@ -539,6 +603,18 @@ def _generate_messages(
                 hash(type_adapter._type), "serialization"
             ]
 
+        bindings: dict[str, dict[str, Any]]
+        if channel._bindings:
+            bindings = {}
+            for type_adapter in channel._bindings.values():
+                binding_type = get_args(type_adapter._type)[1]
+                assert isinstance(binding_type, Binding)
+
+                bindings.setdefault(binding_type.__protocol__, {})[
+                    binding_type.__field_name__
+                ] = field_mapping[hash(type_adapter._type), "serialization"]
+            message["bindings"] = bindings
+
         yield f"{channel.title}Message", message
 
         for channel_message in channel.messages:
@@ -550,6 +626,17 @@ def _generate_messages(
                     hash(type_adapter._type), "serialization"
                 ]
 
+            if channel_message.__bindings__:
+                bindings = {}
+                for type_adapter in channel_message.__bindings__.values():
+                    binding_type = get_args(type_adapter._type)[1]
+                    assert isinstance(binding_type, Binding)
+
+                    bindings.setdefault(binding_type.__protocol__, {})[
+                        binding_type.__field_name__
+                    ] = field_mapping[hash(type_adapter._type), "serialization"]
+                message_message["bindings"] = bindings
+
             yield channel_message.__name__, message_message
 
 
